@@ -136,7 +136,8 @@ my %VCS_EXTRACT = (
     # cvs rootdir followed by optional module name:
     cvs     => sub { return shift =~ /^(.+?)(?:\s+(\S*))?$/;},
     darcs   => sub { return @_;},
-    hg      => sub { return @_;},
+    # hg uri followed by optional -b branchname
+    hg      => sub { return shift =~ /^(.+?)(?:\s+-b\s+(\S*))?$/;},
     # git uri followed by optional "[subdir]", "-b branchname" etc.
     git     =>
       sub { return shift =~ /^(.+?)(?:\s+\[(\S*)\])?(?:\s+-b\s+(\S*))?$/;},
@@ -190,6 +191,7 @@ my %VCS_CANONIFY = (
         ) {
             $_[1] = 'vcs-git-uses-invalid-user-uri';
         }
+        $_[0] =~ s{(https?://.*?\.git)(?:\.git)+$}{$1};
         $_[0] =~ s{https?\Q://git.debian.org/\E(?:git/?)?}
                   {https://anonscm.debian.org/git/};
         $_[0] =~ s{https?\Q://anonscm.debian.org/git/git/\E}
@@ -1183,6 +1185,7 @@ sub run {
                         my ($dep, $pkg_name) = @{$d};
                         my $replacement = $OBSOLETE_PACKAGES->value($pkg_name)
                           // '';
+                        next if $proc->pkg_src eq 'lintian';
                         $replacement = ' => ' . $replacement
                           if $replacement ne '';
                         if (   $pkg_name eq $alternatives[0][0]
