@@ -28,6 +28,7 @@ use Lintian::Relation;
 use Lintian::Tags qw(tag);
 use Lintian::Util qw(
   file_is_encoded_in_non_utf8
+  DCTRL_COMMENTS_AT_EOL
   read_dpkg_control
   strip
 );
@@ -50,8 +51,9 @@ my %KNOWN_SPECIAL_DEPENDS = map { $_ => 1 } qw(
 sub run {
     my ($pkg, $type, $info) = @_;
     my $testsuites = $info->field('testsuite', '');
-    my $control = $info->index('debian/tests/control');
-    my $control_autodep8 = $info->index('debian/tests/control.autodep8');
+    my $control = $info->index_resolved_path('debian/tests/control');
+    my $control_autodep8
+      = $info->index_resolved_path('debian/tests/control.autodep8');
     my $needs_control = 0;
 
     tag 'testsuite-autopkgtest-missing' if ($testsuites !~ /autopkgtest/);
@@ -81,7 +83,7 @@ sub run {
         }
 
         tag 'unnecessary-testsuite-autopkgtest-field'
-          if $info->source_field('testsuite') // '' eq 'autopkgtest';
+          if ($info->source_field('testsuite') // '') eq 'autopkgtest';
 
         tag 'debian-tests-control-and-control-autodep8', $control,
           $control_autodep8
@@ -98,7 +100,12 @@ sub check_control_contents {
     my ($info, $path) = @_;
 
     my (@paragraphs, @lines);
-    if (not eval { @paragraphs = read_dpkg_control($path, 0, \@lines); }) {
+    if (
+        not eval {
+            @paragraphs
+              = read_dpkg_control($path, DCTRL_COMMENTS_AT_EOL, \@lines);
+        }
+    ) {
         chomp $@;
         $@ =~ s/^syntax error at //;
         tag 'syntax-error-in-debian-tests-control', $@;
@@ -193,7 +200,7 @@ sub check_test_file {
     my ($info, $directory, $name, $line) = @_;
     # Special case with "Tests-Directory: ." (see #849880)
     my $path = $directory eq '.' ? $name : "$directory/$name";
-    my $index = $info->index($path);
+    my $index = $info->index_resolved_path($path);
 
     if ($name !~ m{^ [ [:alnum:] \+ \- \. / ]++ $}xsm) {
         tag 'illegal-runtime-test-name', $name,
