@@ -29,7 +29,8 @@ use Scalar::Util qw(blessed);
 use Lintian::Relation;
 use Parse::DebianChangelog;
 
-use Lintian::Util qw(get_file_checksum read_dpkg_control $PKGNAME_REGEX);
+use Lintian::Util
+  qw(get_file_checksum read_dpkg_control open_gz $PKGNAME_REGEX $PKGREPACK_REGEX strip);
 
 =head1 NAME
 
@@ -186,6 +187,105 @@ sub native {
     return $self->{native};
 }
 
+=item files
+
+Returns a reference to a hash containing information about files listed
+in the .changes file.  Each hash may have the following keys:
+
+=over 4
+
+=item name
+
+Name of the file.
+
+=item size
+
+The size of the file in bytes.
+
+=item checksums
+
+A hash with the keys being checksum algorithms and the values themselves being
+hashes containing
+
+=over 4
+
+=item sum
+
+The result of applying the given algorithm to the file.
+
+=item filesize
+
+The size of the file as given in the .changes section relating to the given
+checksum.
+
+=back
+
+=back
+
+Needs-Info requirements for using I<files>: L<Lintian::Collect/field ([FIELD[, DEFAULT]])>
+
+=cut
+
+sub files {
+    my ($self) = @_;
+
+    return $self->{files} if exists $self->{files};
+
+    my %files;
+
+    my $file_list = $self->field('files') || '';
+    local $_;
+    for (split /\n/, $file_list) {
+        strip;
+        next if $_ eq '';
+
+        my ($md5sum,$size,$file) = split(/\s+/o, $_);
+        next if $file =~ m,/,;
+
+        $files{$file}{checksums}{md5} = {
+            'sum' => $md5sum,
+            'filesize' => $size,
+        };
+        $files{$file}{name} = $file;
+        $files{$file}{size} = $size;
+    }
+
+    foreach my $alg (qw(sha1 sha256)) {
+        my $list = $self->field("checksums-$alg") || '';
+        for (split /\n/, $list) {
+            strip;
+            next if $_ eq '';
+
+            my ($checksum, $size, $file) = split(/\s+/o, $_);
+            next if $file =~ m,/,;
+
+            $files{$file}{checksums}{$alg} = {
+                'sum' => $checksum,
+                'filesize' => $size
+            };
+        }
+    }
+
+    $self->{files} = \%files;
+    return $self->{files};
+}
+
+=item repacked
+
+Returns true if the source package has been "repacked" and false otherwise.
+This is determined from the version name containing "dfsg" or similar.
+
+Needs-Info requirements for using I<repacked>: L<Same as field|Lintian::Collect/field ([FIELD[, DEFAULT]])>
+
+=cut
+
+sub repacked {
+    my ($self) = @_;
+    return $self->{repacked} if exists $self->{repacked};
+    $self->{repacked} = $self->field('version', '') =~ $PKGREPACK_REGEX;
+    return $self->{repacked};
+}
+
 =item binaries
 
 Returns a list of the binary and udeb packages listed in the
@@ -299,6 +399,7 @@ sub orig_index {
         # source packages do not have anchored roots as they can be
         # unpacked anywhere...
         'has_anchored_root_dir' => 1,
+        'allow_empty' => 1,
     };
     return $self->_fetch_index_data($load_info, $file);
 }
@@ -427,6 +528,92 @@ sub _load_dctrl {
     return 1;
 }
 
+=item java_info
+
+Returns a hashref containing information about JAR files found in
+source packages, in the form I<file name> -> I<info>, where I<info> is
+a hash containing the following keys:
+
+=over 4
+
+=item manifest
+
+A hash containing the contents of the JAR file manifest. For instance,
+to find the classpath of I<$file>, you could use:
+
+ if (exists $info->java_info->{$file}{'manifest'}) {
+     my $cp = $info->java_info->{$file}{'manifest'}{'Class-Path'};
+     # ...
+ }
+
+NB: Not all jar files have a manifest.  For those without, this will
+value will not be available.  Use exists (rather than defined) to
+check for it.
+
+=item files
+
+A table of the files in the JAR.  Each key is a file name and its value
+is its "Major class version" for Java or "-" if it is not a class file.
+
+=item error
+
+If it exists, this is an error that occurred during reading of the zip
+file.  If it exists, it is unlikely that the other fields will be
+present.
+
+=back
+
+Needs-Info requirements for using I<java_info>: java-info
+
+=cut
+
+sub java_info {
+    my ($self) = @_;
+    return $self->{java_info} if exists $self->{java_info};
+    my $javaf = $self->lab_data_path('java-info.gz');
+    my %java_info;
+    if (!-f $javaf) {
+        # no java-info.gz => no jar files to collect data.  Just
+        # return an empty hash ref.
+        $self->{java_info} = \%java_info;
+        return $self->{java_info};
+    }
+    my $idx = open_gz($javaf);
+    my $file;
+    my $file_list;
+    my $manifest = 0;
+    local $_;
+    while (<$idx>) {
+        chomp;
+        next if m/^\s*$/o;
+
+        if (m#^-- ERROR:\s*(\S.++)$#o) {
+            $java_info{$file}{error} = $1;
+        } elsif (m#^-- MANIFEST: (?:\./)?(?:.+)$#o) {
+            # TODO: check $file == $1 ?
+            $java_info{$file}{manifest} = {};
+            $manifest = $java_info{$file}{manifest};
+            $file_list = 0;
+        } elsif (m#^-- (?:\./)?(.+)$#o) {
+            $file = $1;
+            $java_info{$file}{files} = {};
+            $file_list = $java_info{$file}{files};
+            $manifest = 0;
+        } else {
+            if ($manifest && m#^  (\S+):\s(.*)$#o) {
+                $manifest->{$1} = $2;
+            } elsif ($file_list) {
+                my ($fname, $clmajor) = (m#^([^-].*):\s*([-\d]+)$#);
+                $file_list->{$fname} = $clmajor;
+            }
+
+        }
+    }
+    $self->{java_info} = \%java_info;
+    close($idx);
+    return $self->{java_info};
+}
+
 =item binary_relation (PACKAGE, FIELD)
 
 Returns a L<Lintian::Relation> object for the specified FIELD in the
